@@ -106,20 +106,23 @@ int template_delete( template_t * tmpl )
 
 int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
                      inifile_t * ini, sl4_string_t *tmpdir,
-                     unsigned int row, int item_id, sl4_string_t *errmes )
+                     unsigned int row, int item_id, sl4_string_t *errmes,
+                     int is_rowset )
 {
   unsigned int col;
   cexcel_cell_t *cell;
   const char *address, *text, *sname;
-  sl4_hash_table_t *hash;
+  sl4_hash_table_t *rep_hash, *dataset_hash;
   inifile_column_t *inicol;
 
   sname = cexcel_sheet_get_name( sheet );
 
-  hash = sl4_hash_table_str_new(  );
-  if ( hash == NULL ) {
+  rep_hash = sl4_hash_table_str_new(  );
+  dataset_hash = sl4_hash_table_str_new(  );
+  if ( rep_hash == NULL || dataset_hash == NULL ) {
     ERROR_OUTMEM(  );
   }
+
   /* system.item_id */
   {
     sl4_string_t *itemid_str;
@@ -130,7 +133,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
     if ( sl4_string_sprintf( itemid_str, "%d", item_id ) != 0 ) {
       ERROR_OUTMEM(  );
     }
-    sl4_hash_table_insert( hash, "system.item_id", itemid_str );
+    sl4_hash_table_insert( rep_hash, "system.item_id", itemid_str );
   }
   /* system.now */
   {
@@ -145,7 +148,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
     if ( now_str == NULL ) {
       ERROR_OUTMEM(  );
     }
-    sl4_hash_table_insert( hash, "system.now", now_str );
+    sl4_hash_table_insert( rep_hash, "system.now", now_str );
   }
   /* system.itemtype */
   {
@@ -154,7 +157,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
     if ( itemtype_str == NULL ) {
       ERROR_OUTMEM(  );
     }
-    sl4_hash_table_insert( hash, "system.itemtype", itemtype_str );
+    sl4_hash_table_insert( rep_hash, "system.itemtype", itemtype_str );
   }
   for ( col = 0; col < ini->ncol; col++ ) {
     cell = cexcel_sheet_get_cell( sheet, row, col );
@@ -185,7 +188,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
                               address, sname );
           return -1;
         }
-        _template_assign_fields( hash, inicol->field, bool_str, 0, NULL );
+        _template_assign_fields( rep_hash, inicol->field, bool_str, 0, NULL );
         sl4_string_delete( bool_str );
       }
       break;
@@ -199,7 +202,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
                               address, sname );
           return -1;
         }
-        _template_assign_fields( hash, inicol->field, integer_str, 0,
+        _template_assign_fields( rep_hash, inicol->field, integer_str, 0,
                                  NULL );
         sl4_string_delete( integer_str );
       }
@@ -214,7 +217,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
                               address, sname );
           return -1;
         }
-        _template_assign_fields( hash, inicol->field, string_str, 1,
+        _template_assign_fields( rep_hash, inicol->field, string_str, 1,
                                  NULL );
         sl4_string_delete( string_str );
       }
@@ -257,14 +260,14 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
             ERROR_OUTMEM(  );
           }
         }
-        _template_assign_fields( hash, inicol->field, file_str, 0, NULL );
-        _template_assign_fields( hash, inicol->field, original_file_name,
+        _template_assign_fields( rep_hash, inicol->field, file_str, 0, NULL );
+        _template_assign_fields( rep_hash, inicol->field, original_file_name,
                                  0, ".original_file_name" );
-        _template_assign_fields( hash, inicol->field, mime_type, 0,
+        _template_assign_fields( rep_hash, inicol->field, mime_type, 0,
                                  ".mime_type" );
-        _template_assign_fields( hash, inicol->field, file_size, 0,
+        _template_assign_fields( rep_hash, inicol->field, file_size, 0,
                                  ".file_size" );
-        _template_assign_fields( hash, inicol->field, suffix, 0,
+        _template_assign_fields( rep_hash, inicol->field, suffix, 0,
                                  ".suffix" );
         sl4_string_delete( file_str );
         sl4_string_delete( original_file_name );
@@ -284,7 +287,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
                               address, sname );
           return -1;
         }
-        _template_assign_fields( hash, inicol->field, select_str, 1,
+        _template_assign_fields( rep_hash, inicol->field, select_str, 1,
                                  NULL );
         sl4_string_delete( select_str );
       }
@@ -300,14 +303,78 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
                               address, sname );
           return -1;
         }
-        _template_assign_fields( hash, inicol->field, array_str, 0, NULL );
+        _template_assign_fields( rep_hash, inicol->field, array_str, 0,
+                                 NULL );
         sl4_string_delete( array_str );
       }
       break;
+    case INIFILE_TYPE_DATASET:
+      {
+        strarray_t *dataset_data, *dataset_field;
+        sl4_string_t *dataset_name, *tmp;
+        size_t *dataset_size;
+        dataset_field = strarray_new(  );
+        dataset_name = strarray_get( inicol->options, 0 );
+        tmp = sl4_string_new( sl4_string_get( dataset_name ) );
+        if ( dataset_field == NULL || tmp == NULL ) {
+          ERROR_OUTMEM(  );
+        }
+        strarray_add( dataset_field, tmp );
+        dataset_data =
+          tmpl_dataset_get( text, inicol->options, inicol->is_required );
+        if ( dataset_data == NULL ) {
+          sl4_string_sprintf( errmes,
+                              "Invalid Cell(%s) Found in Sheet \"%s\". Data Type Mismatch.",
+                              address, sname );
+          return -1;
+        }
+        // if dataset hash does not exists, create a new content.
+        dataset_size =
+          ( size_t * ) sl4_hash_table_lookup( dataset_hash,
+                                              sl4_string_get
+                                              ( dataset_name ) );
+        if ( dataset_size == NULL ) {
+          dataset_size = ( size_t * ) malloc( sizeof( size_t ) );
+          if ( dataset_size == NULL ) {
+            ERROR_OUTMEM(  );
+          }
+          *dataset_size = strarray_length( dataset_data );
+          sl4_hash_table_insert( dataset_hash, sl4_string_get( dataset_name ),
+                                 dataset_size );
+        }
+        // compare number of content in dataset array.
+        if ( *dataset_size != strarray_length( dataset_data ) ) {
+          sl4_string_sprintf( errmes,
+                              "Invalid Cell(%s) Found in Sheet \"%s\". Array Size Mismatch.",
+                              address, sname );
+          return -1;
+        }
+        {
+          size_t ii, jj;
+          sl4_string_t *postfix = sl4_string_new( NULL );
+          if ( postfix == NULL ) {
+            ERROR_OUTMEM(  );
+          }
+          for ( ii = 0; ii < strarray_length( dataset_data ); ii++ ) {
+            tmp = strarray_get( dataset_data, ii );
+            for ( jj = 0; jj < strarray_length( inicol->field ); jj++ ) {
+              sl4_string_sprintf( postfix, ".%s#%d",
+                                  sl4_string_get( strarray_get
+                                                  ( inicol->field, jj ) ),
+                                  ii );
+              _template_assign_fields( rep_hash, dataset_field, tmp, 1,
+                                       sl4_string_get( postfix ) );
+            }
+          }
+          sl4_string_delete( postfix );
+        }
+        strarray_delete( dataset_field );
+        strarray_delete( dataset_data );
+      }
+      break;
     case INIFILE_TYPE_RIGHTS:
       {
-        sl4_string_t *rights, *use_cc, *cc_commercial_use,
-          *cc_modification;
+        sl4_string_t *rights, *use_cc, *cc_commercial_use, *cc_modification;
         rights =
           tmpl_rights_get( text, &use_cc, &cc_commercial_use,
                            &cc_modification, inicol->is_required );
@@ -317,12 +384,12 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
                               address, sname );
           return -1;
         }
-        _template_assign_fields( hash, inicol->field, rights, 1, NULL );
-        _template_assign_fields( hash, inicol->field, use_cc, 0,
+        _template_assign_fields( rep_hash, inicol->field, rights, 1, NULL );
+        _template_assign_fields( rep_hash, inicol->field, use_cc, 0,
                                  ".use_cc" );
-        _template_assign_fields( hash, inicol->field, cc_commercial_use, 0,
-                                 ".cc_commercial_use" );
-        _template_assign_fields( hash, inicol->field, cc_modification, 0,
+        _template_assign_fields( rep_hash, inicol->field, cc_commercial_use,
+                                 0, ".cc_commercial_use" );
+        _template_assign_fields( rep_hash, inicol->field, cc_modification, 0,
                                  ".cc_modification" );
         sl4_string_delete( rights );
         sl4_string_delete( use_cc );
@@ -340,7 +407,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
                               address, sname );
           return -1;
         }
-        _template_assign_fields( hash, inicol->field, indexes_str, 0,
+        _template_assign_fields( rep_hash, inicol->field, indexes_str, 0,
                                  NULL );
         sl4_string_delete( indexes_str );
       }
@@ -351,7 +418,7 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
     }
   }
   {
-    /* dump hash data */
+    /* dump replacing hash data */
     size_t bi;
     sl4_hash_bucket_t *buck;
     sl4_string_t *mydata, *mypattern, *tmp, *data;
@@ -364,9 +431,65 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
     if ( data == NULL ) {
       ERROR_OUTMEM(  );
     }
-    if ( hash->nkeys != 0 ) {
-      for ( bi = 0; bi < hash->nbucks; bi++ ) {
-        for ( buck = hash->buckets[bi]; buck; buck = buck->next ) {
+    /* preprocess for rowset */
+    if ( is_rowset ) {
+      strarray_t *matches = strarray_new(  );
+      if ( myonig_match
+           ( "/<{\\s*rowset\\s*}>(.*?)<{\\s*\\/rowset\\s*}>/s",
+             sl4_string_get( data ), matches ) != 0 ) {
+        sl4_string_t *body = strarray_get( matches, 1 );
+        sl4_string_set( data, sl4_string_get( body ) );
+      }
+      strarray_delete( matches );
+    }
+    /* preprocess for dataset */
+    {
+      strarray_t *matches = strarray_new(  );
+      while ( myonig_match
+              ( "/<{\\s*dataset\\s*\\$([^}]+)\\s*}>(.*?)<{\\s*\\/dataset\\s*}>/s",
+                sl4_string_get( data ), matches ) != 0 ) {
+        sl4_string_t *all, *name, *body, *rep_body;
+        size_t ii, pos, *dataset_size;
+        all = strarray_get( matches, 0 );
+        name = strarray_get( matches, 1 );
+        body = strarray_get( matches, 2 );
+        rep_body = sl4_string_new( NULL );
+        dataset_size =
+          ( size_t * ) sl4_hash_table_lookup( dataset_hash,
+                                              sl4_string_get( name ) );
+        for ( ii = 0; ii < *dataset_size; ii++ ) {
+          sl4_string_t *rep_name, *tmp_body1, *tmp_body2;
+          rep_name = sl4_string_new( NULL );
+          sl4_string_sprintf( rep_name, "<{@@%s.\\1#%d}>",
+                              sl4_string_get( name ), ii );
+          tmp_body1 =
+            myonig_replace( "/<{\\s*\\$(.+?)\\s*}>/",
+                            sl4_string_get( rep_name ),
+                            sl4_string_get( body ) );
+          if ( tmp_body1 == NULL ) {
+            ERROR_UNEXPECTED(  );
+          }
+          tmp_body2 =
+            myonig_replace( "/<{@@(.+?)}>/", "<{\\$\\1}>",
+                            sl4_string_get( tmp_body1 ) );
+          if ( tmp_body2 == NULL ) {
+            ERROR_UNEXPECTED(  );
+          }
+          sl4_string_append( rep_body, sl4_string_get( tmp_body2 ) );
+          sl4_string_delete( tmp_body1 );
+          sl4_string_delete( tmp_body2 );
+          sl4_string_delete( rep_name );
+        }
+        pos = sl4_string_find( data, sl4_string_get( all ) );
+        sl4_string_replace( data, pos, sl4_string_length( all ),
+                            sl4_string_get( rep_body ) );
+      }
+      strarray_delete( matches );
+    }
+    /* replace */
+    if ( rep_hash->nkeys != 0 ) {
+      for ( bi = 0; bi < rep_hash->nbucks; bi++ ) {
+        for ( buck = rep_hash->buckets[bi]; buck; buck = buck->next ) {
           mydata = ( sl4_string_t * ) buck->data;
           mykey = ( const char * ) buck->key;
           sl4_string_sprintf( mypattern, "/<{\\$%s}>/", mykey );
@@ -398,10 +521,65 @@ int template_assign( template_t * tmpl, cexcel_sheet_t * sheet,
       sl4_file_delete( xml_file );
     }
     sl4_string_delete( data );
-    sl4_string_delete( mypattern );
   }
   /* free hash box */
-  sl4_hash_table_delete( hash, NULL, NULL );
+  sl4_hash_table_delete( rep_hash, NULL, NULL );
+  {
+    size_t bi, *mem;
+    sl4_hash_bucket_t *buck;
+    if ( dataset_hash->nkeys != 0 ) {
+      for ( bi = 0; bi < dataset_hash->nbucks; bi++ ) {
+        for ( buck = dataset_hash->buckets[bi]; buck; buck = buck->next ) {
+          mem = ( size_t * ) buck->data;
+          free( mem );
+        }
+      }
+    }
+  }
+  sl4_hash_table_delete( dataset_hash, NULL, NULL );
+  return 0;
+}
+
+int template_merge_rowset( template_t * tmpl, sl4_string_t *tempdir, int sid,
+                           int eid, const char *name )
+{
+  sl4_string_t *xml_fname, *body, *data;
+  sl4_file_t *xml_file;
+  size_t siz, blen;
+  int i;
+  char buf[8196];
+  blen = 8195;
+  if ( sid > eid ) {
+    return -1;
+  }
+  body = sl4_string_new( NULL );
+  xml_fname = sl4_string_new( NULL );
+  for ( i = sid; i < eid; i++ ) {
+    sl4_string_sprintf( xml_fname, "/%d.xml", i );
+    xml_file = tempdir_fopen( tempdir, sl4_string_get( xml_fname ), "r" );
+    while ( sl4_file_eof( xml_file ) == 0 ) {
+      siz = sl4_file_read( xml_file, buf, blen );
+      if ( siz > 0 ) {
+        buf[siz] = '\0';
+        sl4_string_append( body, buf );
+      }
+    }
+    sl4_file_close( xml_file );
+    sl4_file_unlink( xml_file );
+    sl4_file_delete( xml_file );
+  }
+  data = myonig_replace( "/<{\\s*rowset\\s*}>.*?<{\\s*\\/rowset\\s*}>/s",
+                         sl4_string_get( body ),
+                         sl4_string_get( tmpl->data ) );
+
+  /* output xml file */
+  sl4_string_sprintf( xml_fname, "/%s.xml", name );
+  xml_file = tempdir_fopen( tempdir, sl4_string_get( xml_fname ), "w" );
+  sl4_file_write( xml_file, sl4_string_get( data ),
+                  sl4_string_length( data ) );
+  sl4_file_close( xml_file );
+  sl4_string_delete( xml_fname );
+  sl4_file_delete( xml_file );
   return 0;
 }
 
